mod explodes;
mod expressions;
mod functions;
mod groupby;
mod groupby_dynamic;
mod predicate_queries;
mod projection_queries;
mod queries;

use polars::prelude::*;

pub(crate) fn fruits_cars() -> DataFrame {
    df!(
            "A"=> [1, 2, 3, 4, 5],
            "fruits"=> ["banana", "banana", "apple", "apple", "banana"],
            "B"=> [5, 4, 3, 2, 1],
            "cars"=> ["beetle", "audi", "beetle", "beetle", "beetle"]
    )
    .unwrap()
}
